% diff_mode.sl -*- mode: Slang; mode: fold -*-
%
%%exts:%% diff, patch

% Unified diff mode viewer and editor for Jed %{{{
% version 1.0.0pre1
% Written by Dino Leonardo Sangoi <dino@trieste.linux.it>
% on a shiny day.
% 
% $Revision: 44 $
% $Author: sangoid $
% $Date: 2005-10-10 10:55:32 +0200 (Mon, 10 Oct 2005) $
%
% It does highlighting a bit like mcedit (the midnight commander editor) does.
% (It uses dfa syntax highlighting).
%
% It also helps editing diffs, usually an unsafe operation. Not that you
% can do it blindly now, but at least this script can rebuild diff markers
% after editing. You can also remove a whole block, adjusting the offsets
% accordingly.
%
% This works only on unified diffs right now, and probably forever (mostly
% because I always use unified diffs).
%
% Usage:
%    put diffmode.sl, treemode.sl, and walk.sl somewhere in your
%    JED_LIBRARY_PATH.
%
%    add to you .jedrc something like:
%        autoload("diff_mode", "diffmode");
%        add_mode_for_extension("diff", "diff");
%        add_mode_for_extension("diff", "patch");
%
% And every file with extension '.diff' or '.patch' will automagically
% get the diff mode applied.
%
% Optionally you can set these variables:
% Diff_Use_Tree   -> whether to use the (experimental) tree mode viewer for
%                    patches containing more than one file. If zero, tree mode
%                    is not used, else the value is the minimum number of files
%                    needed inside a patch to trigger the tree mode viewer.
%                    (default = 4, just a randomly choosen number)
% DiffTree_Indent -> how much to indent lines in tree mode viewer. Default = 8
%
% If you are using jed 0.9.16 or newer, you can also redefine the colors used
% for syntax highlighting, just defining colors for "diff_block", "diff_junk",
% "diff_deleted", "diff_added", "diff_fileold", "diff_filenew", "diff_cmd"
%
% Thanks to:
% - John Davis, for this great editor.
% - Abraham van der Merwe, I took some ideas (and maybe also some code...)
%                          from his diff.sl.
% - Günter Milde, for the long emails we exchanged about jed modes (and for
%                 a lot of slang code I'm using)
%}}}

_debug_info = 1;
_traceback = 1;

require("treemode");
require("walk");
require("keydefs");

% Set Diff_Use_Tree to 0 to avoid using a tree view for the diff.
% Otherwise, set it to the minimun number of files covered by the diff
% needed to use the tree view.
custom_variable("Diff_Use_Tree", 4);
custom_variable("DiffTree_Indent", 8);
% should tree pack single child nodes?
custom_variable("DiffTree_Pack", 1);


%%%% Diff low level helpers %{{{

% return true if I'm looking_at() a header marker.
static define _diff_is_marker()
{
	orelse
	{ looking_at("+++ ") }
	{ looking_at("--- ") }
	{ looking_at("diff ") }
	{ looking_at("Only in ") }
	{ looking_at("Binary files ") };
}

% extract information from a block header line: returns:
% (position on old file, length on old file, position on new file, length on new file)
% [ this heavily uses the stack features of slang ].
static define _diff_parse_block_info(l)
{
	variable pos, len;
	
	% Uhmm, there's a better way to do this?
	if (string_match(l, "@@ \\-\\(\\d+\\),\\(\\d+\\) \\+\\(\\d+\\),\\(\\d+\\) @@", 1) == 0)
		error("malformed block header <"+l+">");
	
	(pos, len) = string_match_nth(1);
	integer(l[[pos:pos+len-1]]);
	(pos, len) = string_match_nth(2);
	integer(l[[pos:pos+len-1]]);
	(pos, len) = string_match_nth(3);
	integer(l[[pos:pos+len-1]]);
	(pos, len) = string_match_nth(4);
	integer(l[[pos:pos+len-1]]);
	%   return (oldpos,oldlen,newpos,newlen)
}

% count the added, deleted, and common lines in a block (is up to the caller
% to narrow a region to a block, and to put the point at the start).
% returns (added lines, deleted lines, common lines)
static define _diff_count_block()
{
	variable countplus = 0;
	variable countminus = 0;
	variable countspace = 0;
	
	while (down(1))
	{
		bol();
		if (eobp())
			break;
		switch(what_char())
		{ case '+' : countplus++; }
		{ case '-' : countminus++; }
		{ countspace++; }
	}
	return (countplus, countminus, countspace);
}

% skip a marker in the given direction (dir = 1 -=> down, dir = -1 -=> up)
static define _diff_skip_header(dir)
{
	variable f = &up_1;
	
	if (dir > 0)
		f = &down_1;
	
	while (bol(), _diff_is_marker())
		!if (@f())
			break;
}
%%%%}}}

%%%% Fast movement functions: these are a bit like c_{top,end}_of_function(). %{{{

% Go to the top of diffed file
define diff_top_of_file()
{
	push_mark();
	% search the diff marker.
	if (bol_bsearch("--- ") == 0) {
		pop_mark(1);
		error("start of file not found.");
	}
	% found, see if there's a diff command line before.
	if (up_1()) {
		bol();
		if (looking_at("diff ") == 0)
			go_down_1();
	}
	pop_mark(0);
}

% Go to the end of diffed file
define diff_end_of_file()
{
	% Skip Junk AND first marker
	go_down_1();
	_diff_skip_header(1);
	
	if (bol_fsearch("--- ") == 0) {
		% I can only assume this is the last file-block.
		eob();
		% if point is on a void line, this is the last line in the file, so
		% skip it.
		if (bolp())
			go_up_1();
	}
	_diff_skip_header(-1);
	go_down_1();
}

% Go to the top of diffed block
define diff_top_of_block()
{
	push_mark();
	if (bol_bsearch("@@ ") == 0) {
		pop_mark(1);
		error("start of block not found.");
	}
	pop_mark(0);
}

% Go to the end of diffed block
define diff_end_of_block()
{
	% Skip Junk
	_diff_skip_header(1);
	go_down_1();
	
	if (bol_fsearch("@@ ") != 0) {
		go_up_1;
		_diff_skip_header(-1);
		go_down_1;
	} else {
		% uhm, maybe this is the last block
		diff_end_of_file();
	}
}
%%%%}}}

%%%% Mark and narrow blocks %{{{
% mark the current file. if 'skipheader' is != 0, don't mark the header.
define diff_mark_file(skipheader)
{
	% if the point is in a header, mark the *next* file, not the previous.
	_diff_skip_header(1);
	
	diff_top_of_file();
	if (skipheader)
		_diff_skip_header(1);
	
	push_visible_mark();
	diff_end_of_file();
}

% mark the current block.
define diff_mark_block(skipheader)
{
	% if the point is in a header, mark the *next* block, not the previous.
	!if (looking_at("@@ "))
		diff_top_of_block();
	
	if (skipheader) {
		go_down_1();
		bol();
	}
	
	push_visible_mark();
	diff_end_of_block();
}

% narrows the current file.
define diff_narrow_to_file(skipheader)
{
	diff_mark_file(skipheader);
	eol();
	!if (eobp())
		go_up_1(); % don't get the last line
	narrow();
}

% narrows the current block.
define diff_narrow_to_block(skipheader)
{
	diff_mark_block(skipheader);
	eol();
	!if (eobp())
		go_up_1(); % don't get the last line
	narrow();
}
%%%%}}}

%%%% Redo diff markers after editing %{{{

% Rewrite a block header after editing it. 'old_off' is the offset to apply 
% to line numbers for 'old' side. 'new_off' is the offset to apply to line
% numbers on the 'new' side.
% returns two values: old_off and new_off modified accordinly to changes on
% this block (so returned values can be passed again to diff_redo_block for
% the next block in this file.
define diff_redo_block(old_off, new_off)
{
	variable countminus = 0;
	variable countplus = 0;
	variable countspace = 0;
	variable oldpos, oldsize, newpos, newsize;
	variable c;
	variable oldheader, newheader;
	
	push_spot();
	diff_narrow_to_block(0);
	bob();
	oldheader = line_as_string();
	(oldpos, oldsize, newpos, newsize) = _diff_parse_block_info(oldheader);
	
	(countplus, countminus, countspace) = _diff_count_block();
	countplus += countspace;
	countminus += countspace;
	newheader = sprintf("@@ -%d,%d +%d,%d @@", oldpos+old_off, countminus, newpos+new_off, countplus);
	flush(sprintf("@@ -%d,%d +%d,%d @@   -->   %s", oldpos, oldsize, newpos, newsize, newheader));
	
	if (strcmp(oldheader, newheader) != 0) {
		bob();
		delete_line();
		insert(newheader + "\n");
	}
	widen();
	pop_spot();
	return (old_off + countminus - oldsize, new_off + countplus - newsize);
}

% redo all blocks of this file, starting at block containing the cursor.
static define _diff_redo_from_here(oldoff, newoff)
{
	variable done;
	
	diff_end_of_block();
	diff_top_of_block();
	
	do {
		(oldoff, newoff) = diff_redo_block(oldoff, newoff);
		diff_end_of_block();
	} while (looking_at("@@ "));
}

% redo all blocks on this file.
define diff_redo_file()
{
	push_spot();
	diff_narrow_to_file(0);
	
	bob();
	_diff_redo_from_here(0, 0);
	widen();
	pop_spot();
}
%%%%}}}

%%%% Remove junk routines %{{{

% Remove "Only in..." lines (from diffs without -N)
define diff_remove_only_lines()
{
	push_spot();
	bob();
	while (bol_fsearch("Only in "))
		delete_line();
	pop_spot();
}

% Remove "Binary files..." lines
define diff_remove_binary_lines()
{
	push_spot();
	bob();
	while (bol_fsearch("Binary files "))
		delete_line();		
	pop_spot();
}

define diff_remove_junk_lines()
{
	diff_remove_only_lines();
	diff_remove_binary_lines();
}
%%%%}}}

%%%% Remove block, optionally rebuilding markers %{{{

define diff_remove_block(redo)
{
	variable countplus, countminus;
	
	push_spot();
	% Ensure that we don't cross a file-block boundary
	diff_narrow_to_file(0);
	pop_spot();
	eol();
	diff_narrow_to_block(0);
	diff_top_of_block();
	% calculate deltas to apply to following blocks
	(countplus, countminus,) = _diff_count_block();
	mark_buffer();
	del_region();
	widen(); % block
	
	if (eobp()) {
		if (what_line <= 4) {
			% This is the last block in this file, delete also the file header
			mark_buffer();
			del_region();
		}
	} else {
		if (redo) {
			push_spot();
			diff_end_of_block();
			_diff_redo_from_here(0, countminus-countplus);
			pop_spot();
		}
	}
	
	widen(); % file
	% All this leaves an extra void line, kill it
	!if (eobp())
		delete_line();
	% place the point at the beginning of the next block
	!if (looking_at("@@ ")) {
		diff_end_of_block();
		diff_top_of_block();
	}
}

% remove a hole file block (far easier than a single block)
define diff_remove_file()
{
	diff_mark_file(0);
	del_region();
}

%%%%}}}

%%%% Standard mode things: keymap, syntax, highlighting %{{{
static variable Diff = "Diff";

define diff_add_saurus_bindings()
{
	definekey("diff_top_of_file", Key_Shift_F11, Diff);
	definekey("diff_end_of_file", Key_Shift_F12, Diff);
	definekey("diff_top_of_block", Key_Ctrl_F11, Diff);
	definekey("diff_end_of_block", Key_Ctrl_F12, Diff);
	definekey("(,) = diff_redo_block(0,0)", Key_F12, Diff);
	definekey("diff_redo_file()", Key_F11, Diff);
	definekey("diff_remove_block(1)", Key_F8, Diff);
	definekey("diff_remove_only_lines()", Key_F9, Diff);
	definekey("diff_jump_to(1)", "", Diff);
	
	%%%% Other Functions
	% diff_mark_file(skipheader);
	% diff_mark_block(skipheader);
	% diff_narrow_to_file(skipheader);
	% diff_narrow_to_block(skipheader);
}

%%% Code at level 0! this will run when the file is evaluated!
!if (keymap_p(Diff)) 
{
	make_keymap(Diff);
	diff_add_saurus_bindings();
}

create_syntax_table(Diff);

#ifdef HAS_DFA_SYNTAX
% DFA syntax table: starting from jed 0.99.16, we can define custom colors
% (lacking this capability this mode fallbacks to (ab-)use keyword colors)
if ( _jed_version >= 9916) 
{
	custom_color("diff_block",   "lightgray", "cyan");      % @@
	custom_color("diff_junk",    "lightgray", "red");       % Only / Binary
	custom_color("diff_deleted", "red",       "black");     % -
	custom_color("diff_added",   "green",     "black");     % +
	custom_color("diff_fileold", "red",       "blue");      % ---
	custom_color("diff_filenew", "green",     "blue");      % +++
	custom_color("diff_cmd",     "lightgray", "blue");      % diff
}
else
{
	%%% For backward compatibility
	set_color("keyword3", "lightgray", "blue");      % @@
	set_color("keyword4", "lightgray", "red");       % Only / Binary
	set_color("keyword5", "red",       "black");     % -
	set_color("keyword6", "green",     "black");     % +
	set_color("keyword7", "red",       "lightgray"); % ---
	set_color("keyword8", "green",     "lightgray"); % +++
	set_color("keyword9", "black",     "lightgray"); % diff
}

static define setup_dfa_callback (name)
{
	if ( _jed_version >= 9916)
	{
		%%% DFA_CACHE_BEGIN %%%
		dfa_enable_highlight_cache("diffmode.dfa", name);
		dfa_define_highlight_rule("^diff .*$",      "diff_cmd",     name);
		dfa_define_highlight_rule("^\\+\\+\\+ .*$", "diff_filenew", name);
		dfa_define_highlight_rule("^\\-\\-\\- .*$", "diff_fileold", name);
		dfa_define_highlight_rule("^\\+.*$",        "diff_added",   name);
		dfa_define_highlight_rule("^\\-.*$",        "diff_deleted", name);
		dfa_define_highlight_rule("^Only .*$",      "diff_junk",    name);
		dfa_define_highlight_rule("^Binary .*$",    "diff_junk",    name);
		dfa_define_highlight_rule("^@@ .*$",        "diff_block",   name);
		dfa_build_highlight_table(name);
		%%% DFA_CACHE_END %%%
	} else {
		% I don't keep a cache here, just to avoid problems when upgrading jed...
		% (and because I can't put more than one pair of DFA_CACHE markers in a
		%  single file).
		% dfa_enable_highlight_cache("diffmode.dfa", name);
		dfa_define_highlight_rule("^diff .*$", "keyword9", name);
		dfa_define_highlight_rule("^\\+\\+\\+ .*$", "keyword8", name);
		dfa_define_highlight_rule("^\\-\\-\\- .*$", "keyword7", name);
		dfa_define_highlight_rule("^\\+.*$", "keyword6", name);
		dfa_define_highlight_rule("^\\-.*$", "keyword5", name);
		dfa_define_highlight_rule("^Only .*$", "keyword4", name);
		dfa_define_highlight_rule("^Binary .*$", "keyword4", name);
		dfa_define_highlight_rule("^@@ .*$", "keyword3", name);
		dfa_build_highlight_table(name);
	}
}

dfa_set_init_callback(&setup_dfa_callback, Diff);

% Highlighting NEEDS dfa for this mode to work.
enable_dfa_syntax_for_mode(Diff);

#endif
%%%%}}}

%%%% Jump to functions %{{{

% try to find the original diffed file, and open it in a new buffer.
% If 'new' is true, open the new version, else open the old version.
define diff_jump_to(new)
{
	variable name, pos;
	variable oldpos, newpos;
	variable marker;
	variable delta = what_line();
	% FIXME: Consider Added and removed lines while computing delta.
	walk_mark_current_position();
	diff_top_of_block();
	delta -= what_line();
	line_as_string(); % on stack
	(oldpos, , newpos, ) = _diff_parse_block_info();
	if (new) {
		marker = "+++ ";
		oldpos = newpos;
	} else
		marker = "--- ";
	oldpos += delta;
	diff_top_of_file();
	() = bol_fsearch(marker);
	name = line_as_string();
	newpos = is_substr(name, "\t") - 2;
	name = name[[4:newpos]];
	read_file(name);
	goto_line(oldpos);
	message(name+":"+string(oldpos));
	walk_goto_current_position();
	walk_store_marked_position();
}

%%%%}}}

%%%% Menu %{{{
%%%%

% Adds to the assoc array 'names' all the file names in this diff
% (taking the 'new' versions, i.e. those with '+++').
% key is the name, value is the line in diff where the file-block starts.
static define _get_files_names(names, func)
{
	variable s;
	
	bob;
	while (bol_fsearch("+++ ")) {
		% grab the name
		goto_column(5);
		push_mark();
		if (ffind("\t") == 0)
			eol();
%			error("invalid diff");
		s = bufsubstr();
		% find the top of this block, to get the correct line number.
		push_spot();
		_diff_skip_header(-1);
		go_down_1();
		names[s] = @func;
		pop_spot();
	}
}

static define get_files_names(names)
{
	_get_files_names(names, &what_line);
}

static define get_files_names_as_marks(names)
{
	_get_files_names(names, &create_user_mark);
}

% Adds to the assoc array 'names' all lines containing
% "Only in " and "Binary files " lines.
% (taking the 'new' versions, i.e. those with '+++').
% key is the name, value is -1 (we don't care where these are)
% TODO: Add those as 'real' files to tree, do some parsing to get the full 
%       name from 'Only in', and pick  only the first (or last?) name in 
%       'Binary files'.
static define get_files_markers(names)
{
	variable s;
	
	bob;
	while (re_fsearch("^[OB][ni][ln][ya][ r][iy][n ][ f]")) {
		if (looking_at("Only in ") or looking_at("Binary files ")) {
			push_mark();
			eol;
			s = bufsubstr();
			names[s] = -1;
		}
	}
}

static define files_popup_callback(popup)
{
	variable names = Assoc_Type[Int_Type];
	
	push_spot();
	get_files_names(names);
	pop_spot();
	
	variable keys = assoc_get_keys(names);
	keys = keys[array_sort(keys)];
	foreach (keys)
	{
		variable key = ();
		variable line = names[key];
		menu_append_item(popup, key, &goto_line, line);
	}
}

public define diff_init_menu(menu)
{
	menu_append_popup(menu, "&File Blocks");
	menu_set_select_popup_callback(strcat(menu, ".&File Blocks"), &files_popup_callback);
	
	menu_append_separator(menu);
	menu_append_item(menu, "&Top of File block", "diff_top_of_file");
	menu_append_item(menu, "&End of File block", "diff_end_of_file");
	menu_append_item(menu, "Top of &Block",      "diff_top_of_block");
	menu_append_item(menu, "End of Bl&ock",      "diff_end_of_block");
	menu_append_separator(menu);
	menu_append_item(menu, "&Mark File block", "diff_mark_file(0)");
	menu_append_item(menu, "M&ark Block",      "diff_mark_block(0)");
	menu_append_separator(menu);
	menu_append_item(menu, "&Goto old file", "diff_jump_to(0)");
	menu_append_item(menu, "Goto &new file", "diff_jump_to(1)");
	menu_append_separator(menu);
	menu_append_item(menu, "&Rebuild block", "(,) = diff_redo_block(0,0)");
	menu_append_item(menu, "Rebuild F&ile",  "diff_redo_file");
}

%%%%}}}

%%%% Tree View %{{{

% This variable contains a mark considered not valid
static variable InvMark;

define difftree_goto_pos()
{
	variable l = what_line, mode;
	variable li;
	
	li = tree_elem_get_data();
	
	if (li == NULL)
		error("diffmode: cannot show this item.");
	
	if (li == InvMark)
		error("diffmode: mark");
	
	walk_mark_current_position();
	
	setbuf(user_mark_buffer(li));
	widen();  % just it case...
	goto_user_mark(li);
	diff_narrow_to_file(0);
	bob;
	
	walk_goto_current_position();
	walk_store_marked_position();
}

define difftree_goto_pos_event(event)
{
	difftree_goto_pos();
	return 1;
}

% this is where all the magic of tree creation is done. Yes, is a
% complex recursive function, if you have a better idea, let me know...
%
% for every row of 'aa' between 'start' and 'end', check the element
% at column 'depth', until finds a different element.
static define _build_section(base, aa, start, end, depth, km);
static define _build_section(base, aa, start, end, depth, km)
{
	variable child = "    ";
	variable i, firststart = start;
	
	flush("depth " +string(depth)+", range ["+string(start)+":"+string(end)+"]");
	for (i = start; i < end; i++) {
		if (aa[start][depth] != aa[i][depth]) {
			% this element is different
			% insert base
			if (base != "") 
				tree_add_element(child, base, InvMark);
			base = aa[start][depth];
			if (depth == length(aa[start])-1)
				% this is the last element in this row, insert it.
				tree_add_element(child, base, km[start]);
			else
				% run on next depth (recursive call)
				_build_section(base, aa, start, i, depth+1, km);
			% turn back to parent, clear base, and set new start.
			() = tree_goto_parent();
			base = "";
			start = i;
		}
	}
	% here we are at the end of our run, handle the last element
	if ((firststart == start) and DiffTree_Pack) {
		% all the block has the same column, pack it with the next
		if (base != "")
			base += "/";
		base += aa[start][depth];
	} else {
		% put base on tree
		if (base != "")
			tree_add_element(child, base, InvMark);
		base = aa[start][depth];
	}
	if (depth == length(aa[start])-1)
		tree_add_element(child, base, km[start]);
	else
		_build_section(base, aa, start, end, depth+1, km);
	% goto parent, but not if I have packed the tree.
	!if ((firststart == start) and DiffTree_Pack)
		() = tree_goto_parent();
}

% fill the tree
static define insert_lines(names, marks)
{
	variable i = 0;
	variable count = 0;
	variable oldnode, node;
	variable aa;
	variable depth = 0;   % 'depth' is the element for every line that 
	                      % _build_section() should handle.
	variable I;           % array sort index.
	variable ks = assoc_get_keys(names); % only the names.
	variable km = assoc_get_values(names);
	
	% sort the names,keeping ks and km in synch
	I = array_sort(ks);
	ks = ks[I];
	km = km[I];
	
	% SLang powerfulness: build an array of arrays containing each 
	% a path element. e.g. if:
	% ks[0] == "/home/sauro/.jedrc"
	% ks[1] == "/home/sauro/.jed/home-lib.sl"
	% this builds:
	% aa[0] = [ "", "home", "sauro", ".jedrc" ]; (length = 4)
	% aa[1] = [ "", "home", "sauro", ".jed", "home-lib.sl" ]; (length = 5)
	% note the first void element, this is because of absolute paths
	% but usually patches use relative ones
	aa = array_map(Array_Type, &strchop, ks, '/', 0);
	% handle absolute paths, setting 'depth' to 1 to tell
	% _build_depth() to skip first column
	if (strcmp(aa[0][0], "") == 0)
		depth = 1;
	%%%   tree_set_user_datatype(Integer_Type);
	InvMark = create_user_mark();
	_build_section("", aa, 0, length(aa), depth, km);
	
	% that's all folks.
}

% this is the main function of tree view creation.
public define difftree()
{
	variable names = Assoc_Type[Mark_Type];
	variable marks = Assoc_Type[Int_Type];
	variable diffbuf, difftreebuf;
	
	get_files_names_as_marks(names);
	get_files_markers(marks);
	
	diffbuf = whatbuf();
	difftreebuf = "*tree of "+diffbuf+"*";
	if (bufferp(difftreebuf))
		delbuf(difftreebuf);
	setbuf(difftreebuf);
	tree_mode();
	tree_set_user_datatype(Mark_Type);
	if (tree_build_kmap("TreeDiff")) {
		definekey("tree_open(1)",  "[", "TreeDiff");
		definekey("tree_close(1)", "]", "TreeDiff");
		definekey("tree_toggle(0)", " ", "TreeDiff");
		definekey("difftree_goto_pos", "^M", "TreeDiff");
	}
	insert_lines(names, marks);
	
	sw2buf(difftreebuf);
	
	tree_set_user_func(&difftree_goto_pos_event, TREE_TOGGLE);
	use_keymap("TreeDiff");
	run_mode_hooks("difftree_mode_hook");
}
%%%%}}}

% returns 0 if at least 'c' files are listed in this patch, > 0
% otherwise.
static define check_num_files(c)
{
	bob();
	while ((not eobp()) and (c > 0)) 
	{
		diff_end_of_file();
		c--;
	}
	bob();
	return c;
}

public define diff_mode()
{
	set_mode(Diff, 0);
	use_syntax_table(Diff);
	use_keymap(Diff);
	mode_set_mode_info(Diff, "init_mode_menu", &diff_init_menu);
	set_buffer_undo(1);
	run_mode_hooks("diff_mode_hook");
	if ((Diff_Use_Tree > 0) and (check_num_files(Diff_Use_Tree) == 0)) 
	{
		flush("Building diff tree...");
		difftree();
		flush("Building diff tree... Done.");
	}
}
